/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * Version: 2.0b 09/20/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */

#include "config.h"
 
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <strings.h>
#include <errno.h>
#include <sys/time.h>
#include <syslog.h>

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SCHED_H
#include <sched.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include "vtun.h"
#include "linkfd.h"
#include "lib.h"

/* Host we are working with. 
 * Used by signal handlers that's why it is global. 
 */
struct vtun_host *lfd_host;

extern int errno;

struct lfd_mod *lfd_mod_head = NULL, *lfd_mod_tail = NULL;

/* Modules functions*/

/* Add module to the end of modules list */
void lfd_add_mod(struct lfd_mod *mod)
{
     if( !lfd_mod_head ){
        lfd_mod_head = lfd_mod_tail = mod;
	mod->next = mod->prev = NULL;
     } else {
        lfd_mod_tail->next = mod;
        mod->prev = lfd_mod_tail;
        mod->next = NULL;
        lfd_mod_tail = mod;
     }
}

/*  Initialize and allocate each module */
int lfd_alloc_mod(struct vtun_host *host)
{
     struct lfd_mod *mod = lfd_mod_head;

     while( mod ){
        if( mod->alloc && (mod->alloc)(host) )
	   return 1; 
	mod = mod->next;
     } 

     return 0;
}

/* Free each module */
int lfd_free_mod(void)
{
     struct lfd_mod *mod = lfd_mod_head;

     while( mod ){
        if( mod->free && (mod->free)() )
	   return 1;
	mod = mod->next;
     } 

     return 0;
}

 /* Run modules down (from head to tail) */
inline int lfd_run_down(int len, char *in, char **out)
{
     struct lfd_mod *mod = lfd_mod_head;
     
     if( !mod )
	*out = in;
     else		   	
        for( ; mod && len > 0; mod = mod->next )
            if( mod->encode ){
               len = (mod->encode)(len,in,out);
               in = *out;
            }
     return len;
}

/* Run modules up (from tail to head) */
inline int lfd_run_up(int len, char *in, char **out)
{
     struct lfd_mod *mod = lfd_mod_tail;
     
     if( !mod )
	*out = in;
     else		   	
        for( ; mod && len > 0; mod = mod->prev )
	    if( mod->decode ){
	       len = (mod->decode)(len,in,out);
               in = *out;
	    }
     return len;
}

/* Check if modules are accepting the data(down) */
inline int lfd_check_down(void)
{
     struct lfd_mod *mod = lfd_mod_head;
     int err = 1;	    
 
     if( mod ){
	for( ; mod && err > 0; mod = mod->next )
           if( mod->avail_encode )
              err = (mod->avail_encode)();
     }
	
     return err;
}

/* Check if modules are accepting the data(up) */
inline int lfd_check_up(void)
{
     struct lfd_mod *mod = lfd_mod_tail;
     int err = 1;	    

     if( mod ){
        for( ; mod && err > 0; mod = mod->prev)
           if( mod->avail_decode )
              err = (mod->avail_decode)();
     }

     return err;
}
/* Function to read/write frames. */

inline int read_frame(int fd, char *buf)
{
     unsigned short len;

     /* Read flags and frame size */ 
     if( read_n(fd,(char *)&len,sizeof(short)) <=0 )
	return -1;

     len = ntohs(len);

     if( len & ~VTUN_FSIZE_MASK ){
	/* If any flags are set return them */
	return len;
     }

     len &= VTUN_FSIZE_MASK;

     if( len > VTUN_FRAME_SIZE )
	return 0;

     /* Read frame */ 
     return read_n(fd,buf,len);
}		

inline int write_frame(int fd, char *buf, short len)
{
     unsigned short nlen = htons(len); 

     if( len & ~VTUN_FSIZE_MASK ){
	 /* If any flags are set write them */
	 return write(fd, &nlen, sizeof(short));
     }    

     /* Write frame size */ 
     if( write_n(fd,(char *)&nlen,sizeof(short)) < 0 )
	return -1;

     /* Write frame */ 
     return write_n(fd,buf,len);
}

/********** Linker *************/

/* Termination flag */
int linker_term = 0;

void sig_term(int sig)
{
     syslog(LOG_INFO,"Closing connection");
     linker_term = 1;
}

/* Dumps statistic on SIGUSR1 */
void sig_usr1(int sig)
{
     syslog(LOG_INFO,"%s unc[ in %ld out %ld ] com[ in %ld out %ld ]",
	lfd_host->host, lfd_host->stat.byte_in, lfd_host->stat.byte_out,
	lfd_host->stat.comp_in, lfd_host->stat.comp_out); 
}    

#define ka_seq(a)  a = (a+1) & (VTUN_FRAME_SIZE-1)

int lfd_linker(void)
{
     int fd1 = lfd_host->rmt_fd;
     int fd2 = lfd_host->loc_fd; 
     int idle = 0, ka_sent = 0;
     struct timeval tv;
     char *buf, *out;
     fd_set fdset;
     int maxfd;

     if( !(buf=malloc(VTUN_FRAME_SIZE)) ){
	syslog(LOG_ERR,"Can't allocate buffer for the linker"); 
        return 1; 
     }

     maxfd = (fd1 > fd2 ? fd1 : fd2) + 1;

     while(1){
        register int len, fl;

	if( linker_term )
           break;

        /* Wait for data */
        FD_ZERO(&fdset);
	FD_SET(fd1, &fdset);
	FD_SET(fd2, &fdset);

 	tv.tv_sec = 60; tv.tv_usec = 0;
	if( (len = select(maxfd, &fdset, NULL, NULL, &tv)) < 0 ){
	   if( errno != EAGAIN && errno != EINTR )
	      break;
	   else
	      continue;
	} 
	
	if( !len ){
	   /* We are idle, lets do something usefull */

	   if( lfd_host->flags & VTUN_KEEPALIVE ){
	      /* Make new keepalive seq num*/
	      ka_seq( ka_sent );
	      write_frame(fd1, NULL, VTUN_KEEP_ALIVE | ka_sent);
	      if( ++idle > 1 ){
	         syslog(LOG_INFO,"Session %s network timeout", lfd_host->host);
		 break;	
	      }
	   }
	   continue;
	}	   

	/* Read frames from network(fd1), decode and pass them to 
         * the local program (fd2) */
	if( FD_ISSET(fd1, &fdset) && lfd_check_up() ){
	   idle = 0; 
	   if( (len=read_frame(fd1,buf)) <= 0 )
	      break;

	   /* Handle frame flags */
	   fl = len & ~VTUN_FSIZE_MASK;
           len = len & VTUN_FSIZE_MASK;
	   switch(fl){
	      case VTUN_KEEP_ALIVE:
		if( len != ka_sent ){
		   /* Reply on keepalive mesg with diffirent seq */
		   write_frame(fd1, NULL, VTUN_KEEP_ALIVE | len);
		}
		continue;
	      case VTUN_CONN_CLOSE:
	        syslog(LOG_INFO,"Connection closed by other side");
		linker_term = 1;
		continue;
	   }   

	   lfd_host->stat.comp_in += len; 
	   if( (len=lfd_run_up(len,buf,&out)) == -1 )
	      break;	
	   if( len && write_n(fd2,out,len) < 0 )
	      break; 
	   lfd_host->stat.byte_in += len; 
	}

	/* Read data from the local program(fd2), encode and pass them to 
         * the network (fd1) */
	if( FD_ISSET(fd2, &fdset) && lfd_check_down() ){
	   if( (len=read(fd2, buf, VTUN_FRAME_SIZE)) < 0 ){
	      if( errno != EAGAIN && errno != EINTR )
	         break;
	      else
		 continue;
	   }
	   if( !len ) break;
	
	   lfd_host->stat.byte_out += len; 
	   if( (len=lfd_run_down(len,buf,&out)) == -1 )
	      break;
	   if( len && write_frame(fd1, out, len) < 0 )
	      break;
	   lfd_host->stat.comp_out += len; 
	}
     }
	
     /* Notify other end about our close */
     write_frame(fd1, NULL, VTUN_CONN_CLOSE);
     free(buf);

     return 0;
}

/* Link remote and local file descriptors */ 

int linkfd(struct vtun_host *host)
{
     struct sigaction sa;
     int old_prio;

     lfd_host = host;
 
     old_prio=getpriority(PRIO_PROCESS,0);
     setpriority(PRIO_PROCESS,0,LINKFD_PRIO);

     /* Build modules stack */
	
     if(host->flags & VTUN_COMPRESS)
	lfd_add_mod(&lfd_zcomp);

     if(host->flags & VTUN_ENCRYPT)
	lfd_add_mod(&lfd_encrypt);

     if(host->flags & VTUN_SHAPE)
	lfd_add_mod(&lfd_shaper);

     if(lfd_alloc_mod(host))
	return 1;

     memset(&sa, 0, sizeof(sa));
     sa.sa_handler=sig_term;
     sigaction(SIGTERM,&sa,NULL);

     sa.sa_handler=sig_usr1;
     sigaction(SIGUSR1,&sa,NULL);
 
     lfd_linker();

     lfd_free_mod();

     setpriority(PRIO_PROCESS,0,old_prio);

     return 0;
}


